package game.worldgen;

import game.biome.Biome;
import game.dimension.Dimension;
import game.rng.NoiseGen;
import game.rng.OctaveGen;
import game.rng.Random;
import game.util.ExtMath;
import game.world.State;
import game.world.WorldServer;

public class GeneratorPerlin implements ChunkGenerator
{
    private static final float[] PARABOLIC = new float[25];
    
    static {
    	for (int x = -2; x <= 2; ++x)
        {
            for (int z = -2; z <= 2; ++z)
            {
                float v = 10.0F / ExtMath.sqrtf((float)(x * x + z * z) + 0.2F);
                PARABOLIC[x + 2 + (z + 2) * 5] = v;
            }
        }
    }
    
    private final NoiseGen lowerNoiseGen;
    private final NoiseGen upperNoiseGen;
    private final NoiseGen mainNoiseGen;
    private final NoiseGen depthNoiseGen;
    private final State block;
    private final State liquid;
    private final float coordinateScale;
    private final float heightScale;
    private final float upperLimitScale;
    private final float lowerLimitScale;
    private final float depthNoiseScaleX;
    private final float depthNoiseScaleZ;
    private final float amplification;
    private final float mainNoiseScaleX;
    private final float mainNoiseScaleY;
    private final float mainNoiseScaleZ;
    private final float baseSize;
    private final float stretchY;
    private final float biomeDepthWeight;
    private final float biomeDepthOffset;
    private final float biomeScaleWeight;
    private final float biomeScaleOffset;
    private final double[] noiseTable = new double[825];
    private final double[] mainNoise = new double[825];
    private final double[] lowerNoise = new double[825];
    private final double[] upperNoise = new double[825];
    private final double[] depthNoise = new double[25];
    private final Biome[] biomes = new Biome[100];
	
//    public GeneratorNew(Random rand, GeneratorSettings settings)
//    {
//    	this(rand, settings.farlands, settings.seaLevel, Blocks.stone.getDefaultState(), 
//    			settings.useLavaSeas ? Blocks.lava.getDefaultState() : Blocks.water.getDefaultState(),
//				settings.coordinateScale, settings.heightScale, settings.upperLimitScale, settings.lowerLimitScale, 
//				settings.depthNoiseScaleX, settings.depthNoiseScaleZ, settings.amplification, settings.mainNoiseScaleX, 
//				settings.mainNoiseScaleY, settings.mainNoiseScaleZ, settings.baseSize, settings.stretchY, settings.biomeDepthWeight, 
//				settings.biomeDepthOffset, settings.biomeScaleWeight, settings.biomeScaleOffset);
//    }
    
    public GeneratorPerlin(Random rand, State block, State liquid, Dimension.GeneratorSettings settings) {
    	this(rand, /* dim.hasFarLands(), dim.getSeaLevel(), */ block, liquid,
    			settings.coordinateScale, settings.heightScale, settings.upperLimitScale, settings.lowerLimitScale, 
				settings.depthNoiseScaleX, settings.depthNoiseScaleZ, settings.amplification, settings.mainNoiseScaleX, 
				settings.mainNoiseScaleY, settings.mainNoiseScaleZ, settings.baseSize, settings.stretchY, settings.biomeDepthWeight, 
				settings.biomeDepthOffset, settings.biomeScaleWeight, settings.biomeScaleOffset);
//    			684.412F, 684.412F, 512.0F, 512.0F, 200.0F, 200.0F, 0.0F, 80.0F, 160.0F, 80.0F, 8.5F, 12.0F, 1.0F, 0.0F, 1.0F, 0.0F);
    }
	
    private GeneratorPerlin(Random rand, /* boolean farlands, int seaLevel, */ State block, State liquid,
    		float coordinateScale, float heightScale, float upperLimitScale, float lowerLimitScale, float depthNoiseScaleX, 
    		float depthNoiseScaleZ, float amplification, float mainNoiseScaleX, float mainNoiseScaleY, float mainNoiseScaleZ, 
    		float baseSize, float stretchY, float biomeDepthWeight, float biomeDepthOffset, float biomeScaleWeight, float biomeScaleOffset)
    {
    	this.lowerNoiseGen = /* farlands ? new OctaveGenOld(rand, 16) : */ new OctaveGen(rand, 16);
        this.upperNoiseGen = /* farlands ? new OctaveGenOld(rand, 16) : */ new OctaveGen(rand, 16);
        this.mainNoiseGen = /* farlands ? new OctaveGenOld(rand, 8) : */ new OctaveGen(rand, 8);
        this.depthNoiseGen = /* farlands ? new OctaveGenOld(rand, 16) : */ new OctaveGen(rand, 16);
        this.liquid = liquid;
        this.block = block;
        this.coordinateScale = coordinateScale;
        this.heightScale = heightScale;
        this.upperLimitScale = upperLimitScale;
        this.lowerLimitScale = lowerLimitScale;
        this.depthNoiseScaleX = depthNoiseScaleX;
        this.depthNoiseScaleZ = depthNoiseScaleZ;
        this.amplification = amplification;
        this.mainNoiseScaleX = mainNoiseScaleX;
        this.mainNoiseScaleY = mainNoiseScaleY;
        this.mainNoiseScaleZ = mainNoiseScaleZ;
        this.baseSize = baseSize;
        this.stretchY = stretchY;
        this.biomeDepthWeight = biomeDepthWeight;
        this.biomeDepthOffset = biomeDepthOffset;
        this.biomeScaleWeight = biomeScaleWeight;
        this.biomeScaleOffset = biomeScaleOffset;
    }
    
	public int getMaximumHeight() {
		return 256;
	}
    
    public void generateChunk(WorldServer world, int x, int z, ChunkPrimer primer)
    {
    	int sea = world.getSeaLevel();
    	world.getBiomeGenerator().getGenBiomes(this.biomes, x * 4 - 2, z * 4 - 2, 10, 10);
        this.genNoisemap(x * 4, 0, z * 4);

        for (int xb = 0; xb < 4; ++xb)
        {
            int x1 = xb * 5;
            int x2 = (xb + 1) * 5;

            for (int zb = 0; zb < 4; ++zb)
            {
                int xz1 = (x1 + zb) * 33;
                int xz2 = (x1 + zb + 1) * 33;
                int xz3 = (x2 + zb) * 33;
                int xz4 = (x2 + zb + 1) * 33;

                for (int yb = 0; yb < 32; ++yb)
                {
                    double mul1 = 0.125D;
                    double m1 = this.noiseTable[xz1 + yb];
                    double m2 = this.noiseTable[xz2 + yb];
                    double m3 = this.noiseTable[xz3 + yb];
                    double m4 = this.noiseTable[xz4 + yb];
                    double n1 = (this.noiseTable[xz1 + yb + 1] - m1) * mul1;
                    double n2 = (this.noiseTable[xz2 + yb + 1] - m2) * mul1;
                    double n3 = (this.noiseTable[xz3 + yb + 1] - m3) * mul1;
                    double n4 = (this.noiseTable[xz4 + yb + 1] - m4) * mul1;

                    for (int ys = 0; ys < 8; ++ys)
                    {
                        double mul2 = 0.25D;
                        double o1 = m1;
                        double o2 = m2;
                        double o3 = (m3 - m1) * mul2;
                        double o4 = (m4 - m2) * mul2;

                        for (int xs = 0; xs < 4; ++xs)
                        {
                            double mul3 = 0.25D;
                            double step = (o2 - o1) * mul3;
                            double val = o1 - step;

                            for (int zs = 0; zs < 4; ++zs)
                            {
                                if ((val += step) > 0.0D)
                                {
                                    primer.set(xb * 4 + xs, yb * 8 + ys, zb * 4 + zs, this.block);
                                }
                                else if (yb * 8 + ys < sea)
                                {
                                    primer.set(xb * 4 + xs, yb * 8 + ys, zb * 4 + zs, this.liquid);
                                }
                            }

                            o1 += o3;
                            o2 += o4;
                        }

                        m1 += n1;
                        m2 += n2;
                        m3 += n3;
                        m4 += n4;
                    }
                }
            }
        }
    }

    private void genNoisemap(int x, int y, int z)
    {
        this.depthNoiseGen.generate(this.depthNoise, x, z, 5, 5, (double)this.depthNoiseScaleX, (double)this.depthNoiseScaleZ, 0.5);
        float scale = this.coordinateScale;
        float height = this.heightScale;
        this.mainNoiseGen.generate(this.mainNoise, x, y, z, 5, 33, 5, (double)(scale / this.mainNoiseScaleX), (double)(height / this.mainNoiseScaleY), (double)(scale / this.mainNoiseScaleZ));
        this.lowerNoiseGen.generate(this.lowerNoise, x, y, z, 5, 33, 5, (double)scale, (double)height, (double)scale);
        this.upperNoiseGen.generate(this.upperNoise, x, y, z, 5, 33, 5, (double)scale, (double)height, (double)scale);
        z = 0;
        x = 0;
        int pos = 0;
        int dpos = 0;

        for (int u = 0; u < 5; ++u)
        {
            for (int v = 0; v < 5; ++v)
            {
                float max = 0.0F;
                float min = 0.0F;
                float sum = 0.0F;
                int range = 2;
                Biome biome = this.biomes[u + 2 + (v + 2) * 10];

                for (int a = -range; a <= range; ++a)
                {
                    for (int b = -range; b <= range; ++b)
                    {
                        Biome biome2 = this.biomes[u + a + 2 + (v + b + 2) * 10];
                        float bmin = this.biomeDepthOffset + biome2.depth * this.biomeDepthWeight;
                        float bmax = this.biomeScaleOffset + biome2.scale * this.biomeScaleWeight;

                        if (this.amplification > 0.0F && bmin > 0.0F)
                        {
                            bmin = 1.0F + bmin * this.amplification;
                            bmax = 1.0F + bmax * this.amplification * 2.0F;
                        }

                        float fact = PARABOLIC[a + 2 + (b + 2) * 5] / (bmin + 2.0F);

                        if (biome2.depth > biome.depth)
                        {
                            fact /= 2.0F;
                        }

                        max += bmax * fact;
                        min += bmin * fact;
                        sum += fact;
                    }
                }

                max = max / sum;
                min = min / sum;
                max = max * 0.9F + 0.1F;
                min = (min * 4.0F - 1.0F) / 8.0F;
                double depth = this.depthNoise[dpos] / 8000.0D;

                if (depth < 0.0D)
                {
                    depth = -depth * 0.3D;
                }

                depth = depth * 3.0D - 2.0D;

                if (depth < 0.0D)
                {
                    depth = depth / 2.0D;

                    if (depth < -1.0D)
                    {
                        depth = -1.0D;
                    }

                    depth = depth / 1.4D;
                    depth = depth / 2.0D;
                }
                else
                {
                    if (depth > 1.0D)
                    {
                        depth = 1.0D;
                    }

                    depth = depth / 8.0D;
                }

                ++dpos;
                double low = (double)min;
                double high = (double)max;
                low = low + depth * 0.2D;
                low = low * (double)this.baseSize / 8.0D;
                double shift = (double)this.baseSize + low * 4.0D;

                for (int w = 0; w < 33; ++w)
                {
                    double damp = ((double)w - shift) * (double)this.stretchY * 128.0D / 256.0D / high;

                    if (damp < 0.0D)
                    {
                        damp *= 4.0D;
                    }

                    double lower = this.lowerNoise[pos] / (double)this.lowerLimitScale;
                    double upper = this.upperNoise[pos] / (double)this.upperLimitScale;
                    double base = (this.mainNoise[pos] / 10.0D + 1.0D) / 2.0D;
                    double noise = (base < 0.0D ? lower : (base > 1.0D ? upper : lower + (upper - lower) * base)) - damp;

                    if (w > 29)
                    {
                        double attn = (double)((float)(w - 29) / 3.0F);
                        noise = noise * (1.0D - attn) + -10.0D * attn;
                    }

                    this.noiseTable[pos] = noise;
                    ++pos;
                }
            }
        }
    }
}
